import minimalmodbus
import serial
from retrying import retry
from epevermodbus.extract_bits import extract_bits


class EpeverChargeController(minimalmodbus.Instrument):
    """Instrument class for Epever Charge Controllers.

    Args:
        * portname (str): port name
        * slaveaddress (int): slave address in the range 1 to 247

    """

    def __init__(self, portname, slaveaddress):
        minimalmodbus.Instrument.__init__(self, portname, slaveaddress)
        self.serial.baudrate = 115200
        self.serial.bytesize = 8
        self.serial.parity = serial.PARITY_NONE
        self.serial.stopbits = 1
        self.serial.timeout = 1
        self.mode = minimalmodbus.MODE_RTU
        self.clear_buffers_before_each_transaction = True

    @retry(wait_fixed=200, stop_max_attempt_number=5)
    def retriable_read_register(
        self, registeraddress, number_of_decimals, functioncode
    ):
        return self.read_register(
            registeraddress, number_of_decimals, functioncode, False
        )

    @retry(wait_fixed=200, stop_max_attempt_number=5)
    def retriable_read_bit(self, registeraddress, functioncode):
        return self.read_bit(registeraddress, functioncode)
    
    #Address range 0x3000
    def get_rated_solar_voltage(self):
        """PV array rated voltage"""
        return self.retriable_read_register(0x3000, 2, 4)

    def get_rated_solar_current(self):
        """PV array rated current"""
        return self.retriable_read_register(0x3001, 2, 4)

    def get_rated_solar_power_l(self):
        """PV array rated power (low 16 bits)"""
        return self.retriable_read_register(0x3002, 2, 4)

    def get_rated_solar_power_h(self):
        """PV array rated power (high 16 bits)"""
        return self.retriable_read_register(0x3003, 2, 4)

    def get_rated_charging_voltage(self):
        """Rated Battery's voltage"""
        return self.retriable_read_register(0x3004, 2, 4)

    def get_rated_charging_current(self):
        """Rated charging current to battery"""
        return self.retriable_read_register(0x3005, 2, 4)

    def get_rated_charging_power_l(self):
        """Rated charging power to battery (low 16 bits)"""
        return self.retriable_read_register(0x3006, 2, 4)

    def get_rated_charging_power_h(self):
        """Charging equipment rated output power (high 16 bits)"""
        return self.retriable_read_register(0x3007, 2, 4)

    def get_charging_mode(self):
        """Charging mode: 0x0001 = PWM"""
        return self.retriable_read_register(0x3008, 0, 4)

    def get_rated_load_current(self):
        """Rated load current"""
        return self.retriable_read_register(0x300E, 2, 4)

    # Address range 0x3100
    def get_solar_voltage(self):
        """PV array input in volts"""
        return self.retriable_read_register(0x3100, 2, 4)

    def get_solar_current(self):
        """PV array input in amps"""
        return self.retriable_read_register(0x3101, 2, 4)

    def get_solar_power_l(self):
        """PV array input power L"""
        return self.retriable_read_register(0x3102, 2, 4)

    def get_solar_power_h(self):
        """PV array input power H"""
        return self.retriable_read_register(0x3103, 2, 4)

    def get_battery_voltage(self):
        """PV array input in volts"""
        return self.retriable_read_register(0x3104, 2, 4)

    def get_battery_current(self):
        """PV array input in amps"""
        return self.retriable_read_register(0x3105, 2, 4)

    def get_battery_power_l(self):
        """PV array input power L"""
        return self.retriable_read_register(0x3106, 2, 4)

    def get_battery_power_h(self):
        """PV array input power H"""
        return self.retriable_read_register(0x3107, 2, 4)

    def get_load_voltage(self):
        """Load output in volts"""
        return self.retriable_read_register(0x310C, 2, 4)

    def get_load_current(self):
        """Load output in amps"""
        return self.retriable_read_register(0x310D, 2, 4)

    def get_load_power_l(self):
        """Load power L"""
        return self.retriable_read_register(0x310E, 2, 4)

    def get_load_power_h(self):
        """Load power H"""
        return self.retriable_read_register(0x310F, 2, 4)
    
    def get_battery_temperature(self):
        """battery temperature"""
        return self.retriable_read_register(0x3110, 2, 4)

    def get_controller_temperature(self):
        """Temperature inside equipment"""
        return self.retriable_read_register(0x3111, 2, 4)

    def get_heatsink_temperature(self):
        """Temperature heatsink"""
        return self.retriable_read_register(0x3112, 2, 4)

    def get_battery_state_of_charge(self):
        """Battery state of charge"""
        return self.retriable_read_register(0x311A, 0, 4)

    def get_remote_battery_temperature(self):
        """The battery temperature measured by remote temperature sensor"""
        return self.retriable_read_register(0x311B, 2, 4)

    def get_battery_real_rated_voltage(self):
        """Battery real rated voltage"""
        return self.retriable_read_register(0x311D, 2, 4)

    # Address range 0x3200
    def get_battery_status(self):
        """Battery status"""
        register_value = self.retriable_read_register(0x3200, 0, 4)

        # D7-4
        temperature_warning_status = {
            0: "NORMAL",
            1: "OVER_TEMP",  # Higher than warning settings
            2: "LOW_TEMP",  # Lower than warning settings
        }[extract_bits(register_value, 4, 0b111)]

        # D3-0
        battery_status = {
            0: "NORMAL",
            1: "OVER_VOLTAGE",
            2: "UNDER_VOLTAGE",
            3: "OVER_DISCHARGE",
            4: "FAULT",
        }[extract_bits(register_value, 0, 0b111)]

        return {
            "wrong_identifaction_for_rated_voltage": bool(
                extract_bits(register_value, 15, 0b1)
            ),
            "battery_inner_resistence_abnormal": bool(
                extract_bits(register_value, 8, 0b1)
            ),
            "temperature_warning_status": temperature_warning_status,
            "battery_status": battery_status,
        }

    def get_battery_status_raw(self):
        """Battery status"""
        return self.retriable_read_register(0x3200, 0, 4)

    def get_charging_equipment_status(self):
        """Charging equipment status"""
        register_value = self.retriable_read_register(0x3201, 0, 4)

        # D15-14
        input_voltage_status = {
            0: "NORMAL",
            1: "NO_INPUT_POWER",
            2: "HIGHER_INPUT",
            3: "INPUT_VOLTAGE_ERROR",
        }[extract_bits(register_value, 14, 0b11)]

        # D3-2
        charging_status = {
            0: "NO_CHARGING",
            1: "FLOAT",
            2: "BOOST",
            3: "EQUALIZATION",
        }[extract_bits(register_value, 2, 0b11)]

        return {
            "input_voltage_status": input_voltage_status,
            "charging_mosfet_is_short_circuit": bool(
                extract_bits(register_value, 13, 0b1)
            ),
            "charging_or_anti_reverse_mosfet_is_open_circuit": bool(
                extract_bits(register_value, 12, 0b1)
            ),
            "anti_reverse_mosfet_is_short_circuit": bool(
                extract_bits(register_value, 11, 0b1)
            ),
            "input_over_current": bool(extract_bits(register_value, 10, 0b1)),
            "load_over_current": bool(extract_bits(register_value, 9, 0b1)),
            "load_short_circuit": bool(extract_bits(register_value, 8, 0b1)),
            "load_mosfet_short_circuit": bool(extract_bits(register_value, 7, 0b1)),
            "disequilibrium_in_three_circuits": bool(
                extract_bits(register_value, 6, 0b1)
            ),
            "pv_input_short_circuit": bool(extract_bits(register_value, 4, 0b1)),
            "charging_status": charging_status,
            "fault": bool(
                extract_bits(register_value, 1, 0b1)
            ),  # this does not seem to be functioning correctly. Fault status is returned when no fault.
            "running": bool(extract_bits(register_value, 0, 0b1)),
        }

    def get_charging_equipment_status_raw(self):
        """Charging equipment status"""
        return self.retriable_read_register(0x3201, 0, 4)

    # Address 0x3300
    def get_maximum_solar_voltage_today(self):
        """Maximum solar voltage today"""
        return self.retriable_read_register(0x3300, 2, 4)

    def get_minimum_solar_voltage_today(self):
        """Minimum battery voltage today"""
        return self.retriable_read_register(0x3301, 2, 4)

    def get_maximum_battery_voltage_today(self):
        """Maximum battery voltage today"""
        return self.retriable_read_register(0x3302, 2, 4)

    def get_minimum_battery_voltage_today(self):
        """Minimum battery voltage today"""
        return self.retriable_read_register(0x3303, 2, 4)

    def get_consumed_energy_today_l(self):
        """Consumed energy today LOW"""
        return self.retriable_read_register(0x3304, 2, 4)
    
    def get_consumed_energy_today_h(self):
        """Consumed energy today HIGH"""
        return self.retriable_read_register(0x3305, 2, 4)

    def get_consumed_energy_month_l(self):
        """Consumed energy month LOW"""
        return self.retriable_read_register(0x3306, 2, 4)
    
    def get_consumed_energy_month_h(self):
        """Consumed energy month HIGH"""
        return self.retriable_read_register(0x3307, 2, 4)

    def get_consumed_energy_year_l(self):
        """Consumed energy year LOW"""
        return self.retriable_read_register(0x3308, 2, 4)
    
    def get_consumed_energy_year_h(self):
        """Consumed energy year HIGH"""
        return self.retriable_read_register(0x3309, 2, 4)

    def get_consumed_energy_total_l(self):
        """Consumed energy total LOW"""
        return self.retriable_read_register(0x330A, 2, 4)
    
    def get_consumed_energy_total_h(self):
        """Consumed energy total HIGH"""
        return self.retriable_read_register(0x330B, 2, 4)
    
    def get_generated_energy_today_l(self):
        """Generated energy today"""
        return self.retriable_read_register(0x330C, 2, 4)

    def get_generated_energy_today_h(self):
        """Generated energy today"""
        return self.retriable_read_register(0x330D, 2, 4)

    def get_generated_energy_month_l(self):
        """Generated energy month"""
        return self.retriable_read_register(0x330E, 2, 4)

    def get_generated_energy_month_h(self):
        """Generated energy month"""
        return self.retriable_read_register(0x330F, 2, 4)

    def get_generated_energy_year_l(self):
        """Generated energy year"""
        return self.retriable_read_register(0x3310, 2, 4)

    def get_generated_energy_year_h(self):
        """Generated energy year"""
        return self.retriable_read_register(0x3311, 2, 4)

    def get_generated_energy_total_l(self):
        """Generated energy total"""
        return self.retriable_read_register(0x3312, 2, 4)

    def get_generated_energy_total_h(self):
        """Generated energy total"""
        return self.retriable_read_register(0x3313, 2, 4)

    def get_carbon_reduction_l(self):
        """Carbon dioxide reduction L"""
        return self.retriable_read_register(0x3314, 2, 4)

    def get_carbon_reduction_h(self):
        """Carbon dioxide reduction H"""
        return self.retriable_read_register(0x3315, 2, 4)

    def get_battery_net_current_l(self):
        """Battery current L"""
        return self.retriable_read_register(0x331B, 2, 4)

    def get_battery_net_current_h(self):
        """Battery current H"""
        return self.retriable_read_register(0x331C, 2, 4)

    def get_battery_temperature(self):
        """Temperature battery"""
        return self.retriable_read_register(0x331D, 2, 4)

    def get_ambient_temperature(self):
        """Temperature ambient"""
        return self.retriable_read_register(0x331E, 2, 4)
    
    # Address range 0x9000
    def get_battery_type(self):
        """Battery type"""
        return {0: "USER_DEFINED", 1: "SEALED", 2: "GEL", 3: "FLOODED"}[
            self.retriable_read_register(0x9000, 2, 3)
        ]

    def get_battery_capacity(self):
        """Battery capacity in amp hours"""
        return self.retriable_read_register(0x9001, 0, 3)

    def get_temperature_compensation_coefficient(self):
        """Temperature compensation coefficient"""
        return self.retriable_read_register(0x9002, 2, 3)

    def get_over_voltage_disconnect_voltage(self):
        """Over voltage disconnect voltage"""
        return self.retriable_read_register(0x9003, 2, 3)

    def get_charging_limit_voltage(self):
        """Charging limit voltage"""
        return self.retriable_read_register(0x9004, 2, 3)

    def get_over_voltage_reconnect_voltage(self):
        """Over voltage reconnect voltage"""
        return self.retriable_read_register(0x9005, 2, 3)

    def get_equalize_charging_voltage(self):
        """Equalize charging voltage"""
        return self.retriable_read_register(0x9006, 2, 3)

    def get_boost_charging_voltage(self):
        """Boost charging voltage"""
        return self.retriable_read_register(0x9007, 2, 3)

    def get_float_charging_voltage(self):
        """Float charging voltage"""
        return self.retriable_read_register(0x9008, 2, 3)

    def get_boost_reconnect_charging_voltage(self):
        """Boost reconnect charging voltage"""
        return self.retriable_read_register(0x9009, 2, 3)

    def get_low_voltage_reconnect_voltage(self):
        """Low voltage reconnect voltage"""
        return self.retriable_read_register(0x900A, 2, 3)

    def get_under_voltage_recover_voltage(self):
        """Under voltage warning recover voltage"""
        return self.retriable_read_register(0x900B, 2, 3)

    def get_under_voltage_warning_voltage(self):
        """Under voltage warning voltage"""
        return self.retriable_read_register(0x900C, 2, 3)

    def get_low_voltage_disconnect_voltage(self):
        """Low voltage disconnect voltage"""
        return self.retriable_read_register(0x900D, 2, 3)

    def get_discharging_limit_voltage(self):
        """Discharging limit voltage"""
        return self.retriable_read_register(0x900E, 2, 3)
    
    def get_datetime_sec_min(self):
        """Get seconds and minutes"""
        return self.retriable_read_register(0x9013, 0, 3)

    def get_datetime_hour_day(self):
        """Get hours and day"""
        return self.retriable_read_register(0x9014, 0, 3)

    def get_datetime_month_year(self):
        """Get month and year"""
        return self.retriable_read_register(0x9015, 0, 3)
    
    def get_datetime(self):
        """Get date"""
        value = self.get_datetime_sec_min()
        seconds = value & 0xFF
        minutes = value >> 8
        
        value = self.get_datetime_hour_day()
        hours = value & 0xFF
        day = value >> 8
        
        value = self.get_datetime_month_year()
        month = value & 0xFF
        year = value >> 8
        
        datetime_string = '20' + str(year).zfill(2) + '-' + str(month).zfill(2) + '-' + str(day).zfill(2) + ' ' + str(hours).zfill(2) + ':' + str(minutes).zfill(2) + ':' + str(seconds).zfill(2)
        return datetime_string

    def get_equalization_charging_cycle(self):
        """Equalization charging cycle"""
        return self.retriable_read_register(0x9016, 0, 3)
    
    def get_battery_temp_upper_warning_limit(self):
        """Battery temperature warning upper limit"""
        return self.retriable_read_register(0x9017, 2, 3)
    
    def get_battery_temp_lower_warning_limit(self):
        """Battery temperature warning lower limit"""
        return self.retriable_read_register(0x9018, 2, 3)

    def get_controller_inner_temp_upper_limit(self):
        """Controller inner temperature upper limit"""
        return self.retriable_read_register(0x9019, 2, 3)

    def get_controller_inner_temp_upper_limit_recover(self):
        """Controller inner temperature upper limit recover"""
        return self.retriable_read_register(0x901A, 2, 3)

    def get_power_comp_temp_upper_limit(self):
        """Power component temperature upper limit"""
        return self.retriable_read_register(0x901B, 2, 3)

    def get_power_comp_temp_upper_limit_recover(self):
        """Power component temperature upper limit recover"""
        return self.retriable_read_register(0x901C, 2, 3)
    
    def get_line_impedance(self):
        """Line impedance"""
        return self.retriable_read_register(0x901D, 2, 3)
    
    def get_night_time_threshold_volt(self):
        """Night TimeThreshold Volt.(NTTV)"""
        return self.retriable_read_register(0x901E, 2, 3)

    def get_light_signal_startup_delay(self):
        """Light signal startup (night) delay time"""
        return self.retriable_read_register(0x901F, 0, 3)

    def get_day_time_threshold_volt(self):
        """Day TimeThreshold Volt.(NTTV)"""
        return self.retriable_read_register(0x9020, 2, 3)    
    
    def get_light_signal_turnoff_delay(self):
        """Light signal turn off(day) delay time"""
        return self.retriable_read_register(0x9021, 0, 3)

    def get_controling_mode(self):
        """Load controling modes"""
        return {
            0: "Manual control",
            1: "Light ON/OFF",
            2: "Light ON + timer",
            3: "Time control",
        }[self.retriable_read_register(0x903D, 0, 3)]

    def get_working_time_lenght_1(self):
        """Working time length 1"""
        return self.retriable_read_register(0x903E, 0, 3)

    def get_working_time_lenght_2(self):
        """Working time length 2"""
        return self.retriable_read_register(0x903F, 0, 3)

    def get_turn_on_time_1_sec(self):
        """Turn on timing 1 seconds"""
        return self.retriable_read_register(0x9042, 0, 3)

    def get_turn_on_time_1_min(self):
        """Turn on timing 1 minutes"""
        return self.retriable_read_register(0x9043, 0, 3)

    def get_turn_on_time_1_hours(self):
        """Turn on timing 1 hours"""
        return self.retriable_read_register(0x9044, 0, 3)

    def get_turn_off_time_1_sec(self):
        """Turn off timing 1 seconds"""
        return self.retriable_read_register(0x9045, 0, 3)

    def get_turn_off_time_1_min(self):
        """Turn off timing 1 minutes"""
        return self.retriable_read_register(0x9046, 0, 3)

    def get_turn_off_time_1_hours(self):
        """Turn off timing 1 hours"""
        return self.retriable_read_register(0x9047, 0, 3)

    def get_turn_on_time_2_sec(self):
        """Turn on timing 2 seconds"""
        return self.retriable_read_register(0x9048, 0, 3)

    def get_turn_on_time_2_min(self):
        """Turn on timing 2 minutes"""
        return self.retriable_read_register(0x9049, 0, 3)

    def get_turn_on_time_2_hours(self):
        """Turn on timing 2 hours"""
        return self.retriable_read_register(0x904A, 0, 3)

    def get_turn_off_time_2_sec(self):
        """Turn off timing 2 seconds"""
        return self.retriable_read_register(0x904B, 0, 3)

    def get_turn_off_time_2_min(self):
        """Turn off timing 2 minutes"""
        return self.retriable_read_register(0x904C, 0, 3)

    def get_turn_off_time_2_hours(self):
        """Turn off timing 2 hours"""
        return self.retriable_read_register(0x904D, 0, 3)

    def get_length_of_night(self):
        """Length of night"""
        return self.retriable_read_register(0x9065, 0, 3)

    def get_battery_rated_voltage(self):
        """Battery rated voltage"""
        return {
            0: "AUTO",
            1: "12V",
            2: "24V",
            3: "36V",
            4: "48V",
            5: "60V",
            6: "110V",
            7: "120V",
            8: "220V",
            9: "240V",
        }[self.retriable_read_register(0x9067, 0, 3)]

    def get_load_timing_control_selection(self):
        """Battery rated voltage"""
        return {
            0: "TIMER ONE",
            1: "TIMER TWO",
        }[self.retriable_read_register(0x9069, 0, 3)]

    def get_default_load_on_off_in_manual_mode(self):
        """Default load On/Off in manual mode"""
        return {0: "OFF", 1: "ON"}[self.retriable_read_register(0x906A, 0, 3)]

    def get_equalize_duration(self):
        """Equalize duration"""
        return self.retriable_read_register(0x906B, 0, 3)

    def get_boost_duration(self):
        """Equalize duration"""
        return self.retriable_read_register(0x906C, 0, 3)

    def get_battery_discharge(self):
        """Battery discharge"""
        return self.retriable_read_register(0x906D, 2, 3)

    def get_battery_charge(self):
        """Battery charge"""
        return self.retriable_read_register(0x906E, 2, 3)

    def get_management_mode(self):
        """Management mode"""
        return {0: "VOLTAGE_COMPENSATION", 1: "SOC"}[
            self.retriable_read_register(0x9070, 0, 3)
        ]
    
    # Address range 0x0
    def get_manual_control(self):
        """Manual control the load"""
        return True if self.retriable_read_bit(0x2, 1) == 1 else False

    def get_enable_load_test_mode(self):
        """Enable load test mode"""
        return True if self.retriable_read_bit(0x5, 1) == 1 else False

    def get_force_load_on(self):
        """Force the load on/off"""
        return True if self.retriable_read_bit(0x6, 1) == 1 else False

    # Address range 0x2000
    def is_device_over_temperature(self):
        """Over temperature inside the device"""
        return True if self.retriable_read_bit(0x2000, 2) == 1 else False

    def is_day(self):
        """Is day time"""
        return not self.is_night()

    def is_night(self):
        """Is night time"""
        return True if self.retriable_read_bit(0x200C, 2) == 1 else False

    ## Write operations ##    
    @retry(wait_fixed=200, stop_max_attempt_number=5)
    def retriable_write_register(
        self, registeraddress, value, number_of_decimals, functioncode
    ):
        self.write_register(
            registeraddress, value, number_of_decimals, functioncode, False
        )

    @retry(wait_fixed=200, stop_max_attempt_number=5)
    def retriable_write_registers(
        self, registeraddress, valuelist
    ):
        self.write_registers(
            registeraddress, valuelist
        )

    @retry(wait_fixed=200, stop_max_attempt_number=5)
    def retriable_write_bit(self, registeraddress, value, functioncode):
        self.write_bit(registeraddress, value, functioncode)

    # Address range 0x9000
    def set_battery_type(self, value):
        """Battery type"""
        self.retriable_write_register(0x9000, value, 0, 6)

    def set_battery_capacity(self, value):
        """Battery capacity in amp hours"""
        self.retriable_write_register(0x9001, value, 0, 6)

    def set_temperature_compensation_coefficient(self, value):
        """Temperature compensation coefficient"""
        self.retriable_write_register(0x9002, value,  2, 6)

    def set_over_voltage_disconnect_voltage(self, value):
        """Over voltage disconnect voltage"""
        self.retriable_write_register(0x9003, value,  2, 6)

    def set_charging_limit_voltage(self, value):
        """Charging limit voltage"""
        self.retriable_write_register(0x9004, value,  2, 6)

    def set_over_voltage_reconnect_voltage(self, value):
        """Over voltage reconnect voltage"""
        self.retriable_write_register(0x9005, value,  2, 6)

    def set_equalize_charging_voltage(self, value):
        """Equalize charging voltage"""
        self.retriable_write_register(0x9006,  value, 2, 6)

    def set_boost_charging_voltage(self, value):
        """Boost charging voltage"""
        self.retriable_write_register(0x9007,  value, 2, 6)

    def set_float_charging_voltage(self, value):
        """Float charging voltage"""
        self.retriable_write_register(0x9008, value,  2, 6)

    def set_boost_reconnect_charging_voltage(self, value):
        """Boost reconnect charging voltage"""
        self.retriable_write_register(0x9009,  value, 2, 6)

    def set_low_voltage_reconnect_voltage(self, value):
        """Low voltage reconnect voltage"""
        self.retriable_write_register(0x900A, value,  2, 6)

    def set_under_voltage_recover_voltage(self, value):
        """Under voltage warning recover voltage"""
        self.retriable_write_register(0x900B, value,  2, 6)

    def set_under_voltage_warning_voltage(self, value):
        """Under voltage warning voltage"""
        self.retriable_write_register(0x900C, value,  2, 6)

    def set_low_voltage_disconnect_voltage(self, value):
        """Low voltage disconnect voltage"""
        self.retriable_write_register(0x900D,  value, 2, 6)

    def set_discharging_limit_voltage(self, value):
        """Discharging limit voltage"""
        self.retriable_write_register(0x900E, value,  2, 6)
    
    def set_datetime_sec_min(self, value):
        """Get seconds and minutes"""
        self.retriable_write_register(0x9013, value,  0, 6)

    def set_datetime_hour_day(self, value):
        """Get hours and day"""
        self.retriable_write_register(0x9014, value,  0, 6)

    def set_datetime_month_year(self, value):
        """Get month and year"""
        self.retriable_write_register(0x9015, value,  0, 6)
    
    def set_equalization_charging_cycle(self, value):
        """Equalization charging cycle"""
        self.retriable_write_register(0x9016, value,  0, 6)
    
    def set_battery_temp_upper_warning_limit(self, value):
        """Battery temperature warning upper limit"""
        self.retriable_write_register(0x9017, value,  0, 6)
    
    def set_battery_temp_lower_warning_limit(self, value):
        """Battery temperature warning lower limit"""
        self.retriable_write_register(0x9018, value,  0, 6)

    def set_controller_inner_temp_upper_limit(self, value):
        """Controller inner temperature upper limit"""
        self.retriable_write_register(0x9019, value,  0, 6)

    def set_controller_inner_temp_upper_limit_recover(self, value):
        """Controller inner temperature upper limit recover"""
        self.retriable_write_register(0x901A, value,  0, 6)

    def set_power_comp_temp_upper_limit(self, value):
        """Power component temperature upper limit"""
        self.retriable_write_register(0x901B, value,  0, 6)

    def set_power_comp_temp_upper_limit_recover(self, value):
        """Power component temperature upper limit recover"""
        self.retriable_write_register(0x901C, value,  0, 6)
    
    def set_line_impedance(self, value):
        """Line impedance"""
        self.retriable_write_register(0x901D, value,  0, 6)
    
    def set_night_time_threshold_volt(self, value):
        """Night TimeThreshold Volt.(NTTV)"""
        self.retriable_write_register(0x901E, value,  0, 6)

    def set_light_signal_startup_delay(self, value):
        """Light signal startup (night) delay time"""
        self.retriable_write_register(0x901F, value,  0, 6)

    def set_day_time_threshold_volt(self, value):
        """Day TimeThreshold Volt.(NTTV)"""
        self.retriable_write_register(0x9020, value,  0, 6)    
    
    def set_light_signal_turnoff_delay(self, value):
        """Light signal turn off(day) delay time"""
        self.retriable_write_register(0x9021, value,  0, 6)

    def set_controling_mode(self, value):
        """Load controling modes"""
        self.retriable_write_register(0x903D, value,  0, 6)

    def set_working_time_lenght_1(self, value):
        """Working time length 1"""
        self.retriable_write_register(0x903E, value,  0, 6)

    def set_working_time_lenght_2(self, value):
        """Working time length 2"""
        self.retriable_write_register(0x903F, value,  0, 6)

    def set_turn_on_time_1_sec(self, value):
        """Turn on timing 1 seconds"""
        self.retriable_write_register(0x9042, value,  0, 6)

    def set_turn_on_time_1_min(self, value):
        """Turn on timing 1 minutes"""
        self.retriable_write_register(0x9043, value,  0, 6)

    def set_turn_on_time_1_hours(self, value):
        """Turn on timing 1 hours"""
        self.retriable_write_register(0x9044, value,  0, 6)

    def set_turn_off_time_1_sec(self, value):
        """Turn off timing 1 seconds"""
        self.retriable_write_register(0x9045, value,  0, 6)

    def set_turn_off_time_1_min(self, value):
        """Turn off timing 1 minutes"""
        self.retriable_write_register(0x9046, value,  0, 6)

    def set_turn_off_time_1_hours(self, value):
        """Turn off timing 1 hours"""
        self.retriable_write_register(0x9047, value,  0, 6)

    def set_turn_on_time_2_sec(self, value):
        """Turn on timing 2 seconds"""
        self.retriable_write_register(0x9048, value,  0, 6)

    def set_turn_on_time_2_min(self, value):
        """Turn on timing 2 minutes"""
        self.retriable_write_register(0x9049, value,  0, 6)

    def set_turn_on_time_2_hours(self, value):
        """Turn on timing 2 hours"""
        self.retriable_write_register(0x904A,  value, 0, 6)

    def set_turn_off_time_2_sec(self, value):
        """Turn off timing 2 seconds"""
        self.retriable_write_register(0x904B,  value, 0, 6)

    def set_turn_off_time_2_min(self, value):
        """Turn off timing 2 minutes"""
        self.retriable_write_register(0x904C,  value, 0, 6)

    def set_turn_off_time_2_hours(self, value):
        """Turn off timing 2 hours"""
        self.retriable_write_register(0x904D,  value, 0, 6)

    def set_length_of_night(self, value):
        """Length of night"""
        self.retriable_write_register(0x9065,  value, 0, 6)

    def set_battery_rated_voltage(self, value):
        """Battery rated voltage"""
        self.retriable_write_register(0x9067, value,  0, 6)

    def set_load_timing_control_selection(self, value):
        """Battery rated voltage"""
        self.retriable_write_register(0x9069, value,  0, 6)

    def set_default_load_on_off_in_manual_mode(self, value):
        """Default load On/Off in manual mode"""
        self.retriable_write_register(0x906A,  value, 0, 6)

    def set_equalize_duration(self, value):
        """Equalize duration"""
        self.retriable_write_register(0x906B,  value, 0, 6)

    def set_boost_duration(self, value):
        """Equalize duration"""
        self.retriable_write_register(0x906C, value, 0, 6)

    def set_battery_discharge(self, value):
        """Battery discharge"""
        self.retriable_write_register(0x906D, value, 0, 6)

    def set_battery_charge(self, value):
        """Battery charge"""
        self.retriable_write_register(0x906E, value, 0, 6)

    def set_management_mode(self, value):
        """Management mode"""
        self.retriable_write_register(0x9070, value, 0, 6)
        
    def write_date(self, datetime):
        """Write date and time"""
        value1 = datetime['seconds'] + 256*datetime['minutes']
        value2 = datetime['hours'] + 256*datetime['day']
        value3 = datetime['month'] + 256*datetime['year']
        datetime_list = [value1, value2, value3]
        self.retriable_write_registers(36883, datetime_list);

    def switch_load_on(self):
        """Switch LOAD on"""
        self.retriable_write_bit(0x2, 1, 5)
    
    def switch_load_off(self):
        """Switch LOAD off"""
        self.retriable_write_bit(0x2, 0, 5)
