# Copyright 2021 Marcel Konstapel
# https://www.meezenest.nl/mees
#
# This file is part of SolarGeneratorDatalogger.
# A set of programs and scripts to log all the available data from a self build solar generator.
# It runs on a Raspberry Pi Zero 1 W with Raspbian
#
# It logs data from a Liontron battery (via Bluetooth) and an EPEVER solar charger (via RS485-USB).
# The data is analyzed and stored in a file. The data is also available via a web browser over Wifi
# (Raspberry Pi acts as access point).
# Future plan is to implement LoRa so the data can be accessed remotely.
#
# Program in configured via the file "config.json"
#
# Example of config_file:
#
#    {
#        "liontron_mac": "AA:BB:CC:DD:EE:FF",
#        "liontron_bms_data_file": "liontron_bms_data.json"
#    }
#
# A lot of effort is made to catch all possible communication errors, so the program never hangs.
# When an error occurs, the program exits and has probably not written any data to  'liontron_bms_data_file'
# But at least we did not crash! Check the date of the data file to make sure the data is written or not.
#
# Example of liontron_bms_data_file (JSON format):
#
#   {
#       "total_voltage": 13.19,
#       "current": 0.0,
#       "remain_cap": 44.16,
#       "typ_cap": 64.56,
#       "cycles": 1,
#       "prod_date": 10527,
#       "balance_list": 0,
#       "prot_status": 0,
#       "softwar_version": 37,
#       "remain_cap_percent": 68,
#       "fet_bits": 3,
#       "cell_count": 4,
#       "ntc_count": 2,
#       "temp_1": 16.0,
#       "temp_2": 15.8,
#       "power": 0.0,
#       "voltage_01": 3.298,
#       "voltage_02": 3.298,
#       "voltage_03": 3.298,
#       "voltage_04": 3.297,
#       "system_datetime": "2021-11-09 15:12:25"
#   }
#
# This code is based on:
#   https://github.com/sw-home/FHEM-BluetoothSmartBMS
#   https://github.com/stefanandres/bms-monitoring-stack
#   https://blog.ja-ke.tech/2020/02/07/ltt-power-bms-chinese-protocol.html
#
# Thanks for sharing your knowledge!
#
# V0.1 2021-11-08
#   - BLE communication
#   - Wake up BMS from sleep mode (workaround for bug -undocumented feature?- in Liontron batteries)
#   - BMS data decoding
#   - BMS checking length of data
#   - BMS CRC check
#   - Exits program when gatt detects an error
#   - Configurable via config.json
#   - Writes BMS data to file in JSON format.
#
# V0.2 2021-11-09
#   - Timeout implemented on gatt process, so the program no longer freezes when gatt gets stuck in
#     infinite waiting loop when BLE communication stops unexpectedly.
#
# V0.3 2021-11-10
#   - Removed sys.exit() from BLE managers error handlers. Now it relies completely on the time-out from V0.2
#     Now the program alway exits from the main loop, as it should.
#   - JSON file format for data-file now conforms to standard JSON format.
#   - Implemented error handlers for reading configuration file.
#
# V0.4 2021-11-15
#   - system_date_time is stored in json file instead of date and time as seperate values
#
# This Python3 program gathers the data from the Liontron BMS over Bluetooth BLE.
#
#    SolarGeneratorDatalogger is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SolarGeneratorDatalogger is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SolarGeneratorDatalogger.  If not, see <https://www.gnu.org/licenses/>.
#

import gatt
import json
import sys
import time

from json.decoder import JSONDecodeError
from datetime import datetime
from itertools import count
from multiprocessing import Process

manager = gatt.DeviceManager(adapter_name='hci0')

class AnyDevice(gatt.Device):
    def connect_succeeded(self):
        super().connect_succeeded()
        print("[%s] Connected" % (self.mac_address))

    def connect_failed(self, error):
        super().connect_failed(error)
        print("[%s] Connection failed: %s" % (self.mac_address, str(error)))
        # The gatt-manager stops here and waits in an endless loop, waiting for Godot
        # It relies on the time-out functionallity of the main loop to end the process.
        
        # Or alternative we can force the program to stop. This is not a very elegant solution, but it works.
        #sys.exit("Program ended  due to an error.")
        
    def disconnect_succeeded(self):
        super().disconnect_succeeded()
        print("[%s] Disconnected" % (self.mac_address))
        self.manager.stop()

    def services_resolved(self):
        super().services_resolved()

        device_information_service = next(
            s for s in self.services
            if s.uuid == '0000ff00-0000-1000-8000-00805f9b34fb')

        self.bms_read_characteristic = next(
            c for c in device_information_service.characteristics
            if c.uuid == '0000ff01-0000-1000-8000-00805f9b34fb')

        self.bms_write_characteristic = next(
            c for c in device_information_service.characteristics
            if c.uuid == '0000ff02-0000-1000-8000-00805f9b34fb')

        print("BMS found")
        
        self.bms_read_characteristic.enable_notifications()

    def characteristic_enable_notifications_succeeded(self, characteristic):
        super().characteristic_enable_notifications_succeeded(characteristic)
        print("BMS request hardware info")
        self.response=bytearray()
        self.rawdat={}
        self.wakeup_flag=True
        
        # Bug (or feature) in battery: BMS does not respond immediately. It is probably in deep sleep. Only after a couple of writes, it starts to respond.
        # So, let's wake the BMS up by sending a couple of "BMS HARDWARE INFO" commands to it.
        # When it comes to life we will get a characteristic_value_updated-event
        # From there, we will start the actual communication.
        self.bms_write_characteristic.write_value(bytes([0xDD,0xA5,0x05,0x00,0xFF,0xFB,0x77]));
        self.bms_write_characteristic.write_value(bytes([0xDD,0xA5,0x05,0x00,0xFF,0xFB,0x77]));
        self.bms_write_characteristic.write_value(bytes([0xDD,0xA5,0x05,0x00,0xFF,0xFB,0x77]));
        self.bms_write_characteristic.write_value(bytes([0xDD,0xA5,0x05,0x00,0xFF,0xFB,0x77]));        
        self.bms_write_characteristic.write_value(bytes([0xDD,0xA5,0x05,0x00,0xFF,0xFB,0x77]));
        self.bms_write_characteristic.write_value(bytes([0xDD,0xA5,0x05,0x00,0xFF,0xFB,0x77]));
        self.bms_write_characteristic.write_value(bytes([0xDD,0xA5,0x05,0x00,0xFF,0xFB,0x77]));
        self.bms_write_characteristic.write_value(bytes([0xDD,0xA5,0x05,0x00,0xFF,0xFB,0x77]));
        
    def characteristic_enable_notifications_failed(self, characteristic, error):
        super.characteristic_enable_notifications_failed(characteristic, error)
        print("BMS notification failed:",error)
        # The gatt-manager stops here and waits in an endless loop, waiting for Godot
        # It relies on the time-out functionallity of the main loop to end the process.
        
        # Or alternative we can force the program to stop. This is not a very elegant solution, but it works.
        #sys.exit("Program ended  due to an error.")
        
    # CRC check for BMS data string
    def verify(self, packet):
        data = packet[2:-3]
        check = packet[-3:-1]
        crc=0x10000
        for i in data:
            crc = crc-int(i)
        crc_b = crc.to_bytes(2, byteorder='big')
        return check == crc_b

    def characteristic_value_updated(self, characteristic, value):
        # The first time (or couple of times) this event is triggered, it could be from trying to wake up the BMS (see "characteristic_enable_notifications_succeeded"-event
        # We expect the answering string to be "dd05000a30383230353530343235fdf777". After receiving this string, we send one "data request" to gather the BMS data.
        self.hardware_info=bytearray([0xdd,0x05,0x00,0x0a,0x30,0x38,0x32,0x30,0x35,0x35,0x30,0x34,0x32,0x35,0xfd,0xf7,0x77])
        print("BMS answering")
        self.response+=value
        # BMS answers, lets check if it is a wake-up call or a proper data read
        if (self.response.endswith(b'w')):
            print("BMS answer:", self.response.hex())
            # It's an answer to the wake up call. The BMS is up and running! We can now start communicating.
            # Because the BMS can answer multiple times from the wake up call (as we send multiple wake up calls in a row),
            # we make sure we only send one data request for 'basic info'.
            if self.response[-17:] == self.hardware_info:
                print("BMS wake-up call.")
                # Clear buffer
                self.response=bytearray()
                # Now we can try to get some actual data from the BMS
                # But send request for 'basic info' only one time.
                if (self.wakeup_flag):
                    print("BMS request basic info")
                    self.bms_write_characteristic.write_value(bytes([0xDD,0xA5,0x03,0x00,0xFF,0xFD,0x77]));
                    self.wakeup_flag=False
            # Now we are talking! All incoming messages should be data, but lets check it anyway.
            # 0xDD, 0x03 is a response to 'basic info'
            # position 2 is always 0, position 3 is the length of the data, which is the length of the string minus 7 (the header, the checksum and the 0x77 tail)
            # Check if header is 0xDD, 0x03 AND check if length is correct AND CRC check
            elif self.response[0:2] == bytes([0xDD, 0x03]) and self.response[3] == len(self.response)-7 and self.verify(self.response[:]):
                print("BMS basic info.") 
                # Decode the string we got from the battery
                self.response=self.response[4:]
                self.rawdat['total_voltage']=int.from_bytes(self.response[0:2], byteorder = 'big',signed=True)/100.0
                self.rawdat['current']=int.from_bytes(self.response[2:4], byteorder = 'big',signed=True)/100.0
                self.rawdat['remain_cap']=int.from_bytes(self.response[4:6], byteorder='big', signed=True)/100
                self.rawdat['typ_cap']=int.from_bytes(self.response[6:8], byteorder='big', signed=True)/100
                self.rawdat['cycles']=int.from_bytes(self.response[8:10], byteorder='big', signed=True)
                self.rawdat['prod_date']=int.from_bytes(self.response[10:12],byteorder = 'big',signed=False)
                self.rawdat['balance_list']=int.from_bytes(self.response[12:16],byteorder = 'big',signed=False)
                self.rawdat['prot_status']=int.from_bytes(self.response[16:18],byteorder = 'big',signed=False)
                self.rawdat['softwar_version']=int.from_bytes(self.response[18:19],byteorder = 'big',signed=False)
                self.rawdat['remain_cap_percent']=int.from_bytes(self.response[19:20],byteorder = 'big',signed=False)
                self.rawdat['fet_bits']=int.from_bytes(self.response[20:21],byteorder = 'big',signed=False)
                self.rawdat['cell_count']=int.from_bytes(self.response[21:22],byteorder = 'big',signed=False)
                self.rawdat['ntc_count']=int.from_bytes(self.response[22:23],byteorder = 'big',signed=False)
                
                for i in range(int.from_bytes(self.response[22:23],'big')): # read temperatures
                    self.rawdat['temp_{0:0=1}'.format(i+1)]=(int.from_bytes(self.response[23+i*2:i*2+25],'big')-2731)/10

                self.rawdat['power']=round(self.rawdat['total_voltage']*self.rawdat['current'], 1)
                
                print("BMS request voltages")
                self.response=bytearray()
                self.bms_write_characteristic.write_value(bytes([0xDD,0xA5,0x04,0x00,0xFF,0xFC,0x77]));
                
            # 0xDD, 0x04 is a response to 'voltage info'
            # position 2 is always 0, position 3 is the length of the data, which is the length of the string minus 7 (the header, the checksum and the 0x77 tail)
            # Check if header is 0xDD, 0x03 AND check if length is correct AND CRC check
            elif self.response[0:2] == bytes([0xDD, 0x04]) and self.response[3] == len(self.response)-7 and self.verify(self.response[:]):
                print("BMS cell voltages.") 
                # Decode the string we got from the battery
                self.response=self.response[4:]
                packVolts=0
                for i in range(int(len(self.response)/2)-1):
                    cell=int.from_bytes(self.response[i*2:i*2+2], byteorder = 'big')/1000
                    self.rawdat['voltage_{0:0=2}'.format(i+1)]=cell
                    packVolts+=cell
                print("BMS chat ended")
                self.disconnect()
                # Save the BMS values and current time to a file for further processing by another program
                self.rawdat['system_datetime'] = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
                print (json.dumps(self.rawdat, indent=1, sort_keys=False))
                with open(obj['liontron_bms_data_file'], 'w') as f:
                    json.dump(self.rawdat, f, indent=1)                    
            
            # Oops, something went wrong. Lets try again    
            else:
                print("Fallback. Retry")
                # Clear buffer
                self.response=bytearray()
                # Request basic info
                self.bms_write_characteristic.write_value(bytes([0xDD,0xA5,0x03,0x00,0xFF,0xFD,0x77]));

    def characteristic_write_value_failed(self, characteristic, error):
        print("BMS write failed:",error)
        # The gatt-manager stops here and waits in an endless loop, waiting for Godot
        # It relies on the time-out functionallity of the main loop to end the process.
        
        # Or alternative we can force the program to stop. This is not a very elegant solution, but it works.
        #sys.exit("Program ended due to an error.")


if (len(sys.argv)<2):
    print("Usage: liontron-bms-reader.py <config-file>")
else:

    # read configuration file
    try:
        with open(sys.argv[1], 'r') as myfile:
            data=myfile.read()
    except FileNotFoundError:
        print("Configuration file <"+ sys.argv[1] +"> not found.")
        sys.exit();

    # parse configuration file
    try:
        obj = json.loads(data)
        tmp1 = obj['liontron_mac']
        tmp = obj['liontron_bms_data_file']
    except JSONDecodeError as e:
        print("Error decoding configuration file. Not JSON format?")
        sys.exit()
    except TypeError as e:
        print("Error decoding configuration file: JSON objects not str, bytes or bytearray.")
    except KeyError:
        print("Error reading configuration file: needed fields not found.")
        sys.exit()

    # show values from config file
    print("Using BLE address: " + str(obj['liontron_mac']))
    print("Saving BMS data to: " + str(obj['liontron_bms_data_file']))

    # Do Bluetooth things
    device = AnyDevice(mac_address=obj['liontron_mac'], manager=manager)
    device.connect()
    
    # Do not call manager.run() directly as it has no timeout functionality. Use multiprocessing instead.
    #manager.run()
   
    # Counter is an infinite iterator
    counter = count(0)

    # Start Bluetooth communication as process and let it time out in 10 seconds.
    # The gatt library does not offer a timeout. When an error in the BLE communication occurs,
    # the manager.run() process is stuck in an infinite loop. That is bad!
    # Starting the manager.run() as a multiprocessing process, the multiprocessing-library
    # can let it time-out.
    ble_process = Process(target=manager.run, name='BLE manager')
    ble_process.start()
    ble_process.join(timeout=10)
    ble_process.terminate()
    
    if ble_process.exitcode is None:
        print("BLE communication timeout.")
    
