# Copyright 2021 Marcel Konstapel
# https://www.meezenest.nl/mees
#
# This file is part of SolarGeneratorDatalogger.
# A set of programs and scripts to log all the available data from a self build solar generator.
# It runs on a Raspberry Pi Zero 1 W with Raspbian
#
# It logs data from a EPEVER Soalrge Charge Controller (via Bluetooth) and an EPEVER solar charger (via RS485-USB).
# The data is analyzed and stored in a file. The data is also available via a web browser over Wifi
# (Raspberry Pi acts as access point).
# Future plan is to implement LoRa so the data can be accessed remotely.
#
# This program processes the csv file from the previous day
# It calculates various values and appends this as a row to the big_data file
#
# Usage of program:
#
#  process_daily_data config.json
#
# V0.1 2021-11-15
#   - First working program
#
#    SolarGeneratorDatalogger is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SolarGeneratorDatalogger is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SolarGeneratorDatalogger.  If not, see <https://www.gnu.org/licenses/>.
#
import json
import sys
import time
import getopt
import csv
import os.path
import pandas as pd

from json.decoder import JSONDecodeError
from datetime import datetime

if (len(sys.argv)<2):
    print("Usage: process_daily_data.py <config-file>")
else:

    # read configuration file
    try:
        with open(sys.argv[1], 'r') as myfile:
            data=myfile.read()
    except FileNotFoundError:
        print("Configuration file <"+ sys.argv[1] +"> not found.")
        sys.exit();

    # parse configuration file
    try:
        obj = json.loads(data)
        tmp = obj['csv_bms_data_file']
        tmp1 = obj['csv_solar_data_file']
        tmp2 = obj['csv_bms_bigdata_file']
        tmp2 = obj['csv_solar_bigdata_file']        
    except JSONDecodeError as e:
        print("Error decoding configuration file. Not JSON format?")
        sys.exit()
    except TypeError as e:
        print("Error decoding configuration file: JSON objects not str, bytes or bytearray.")
    except KeyError:
        print("Error reading configuration file: needed fields not found.")
        sys.exit()

    # show values from config file
    print("Reaidng BMS csv data from    : " + str(obj['csv_bms_data_file']))
    print("Reading solar csv data from  : " + str(obj['csv_solar_data_file']))
    print("Writing BMS csv big data to  : " + str(obj['csv_bms_bigdata_file']))
    print("Writing solar csv big data to: " + str(obj['csv_solar_bigdata_file']))
    
    # open BMS data file and extract values
    try:
        data = pd.read_csv(obj['csv_bms_data_file'])
    except FileNotFoundError:
        print("Input file <"+ obj['csv_bms_data_file'] +"> not found.")
        sys.exit();

#system_datetime,total_voltage,current,remain_cap,typ_cap,cycles,prod_date,balance_list,prot_status,softwar_version,remain_cap_percent,fet_bits,cell_count,ntc_count,temp_1,temp_2,power,voltage_01,voltage_02,voltage_03,voltage_04
    iso_date = datetime.fromisoformat(data.iloc[0]['system_datetime'])
    
    csv_data = []
    csv_data.insert(0,iso_date.date())
    csv_data.insert(1, data['total_voltage'].max())
    csv_data.insert(2, data['total_voltage'].min())    
    csv_data.insert(3, data['current'].max())
    csv_data.insert(4, data['current'].min())    
    csv_data.insert(5, data['remain_cap'].max())
    csv_data.insert(6, data['remain_cap'].min())

    # Get delta of energy stored
    csv_data.insert(7, data.iloc[-1]['remain_cap'] - data.iloc[0]['remain_cap'])
    csv_data.insert(8, data['typ_cap'].max())
    csv_data.insert(9, data['typ_cap'].min())    
    csv_data.insert(10, data['cycles'].max())
    csv_data.insert(11, data['prod_date'].max())
    
    # Get all bits from today
    balance_list_orred = 0
    for index, row in data.iterrows():
        balance_list_orred = balance_list_orred | row['balance_list']
    csv_data.insert(12, balance_list_orred)
    
    # Get all bits from today
    prot_status_orred = 0
    for index, row in data.iterrows():
        prot_status_orred = prot_status_orred | row['balance_list']
    csv_data.insert(13, prot_status_orred)
    
    csv_data.insert(14, data['softwar_version'].max())        
    csv_data.insert(15, data['remain_cap_percent'].max())    
    csv_data.insert(16, data['remain_cap_percent'].min())

    # Get all bits from today
    fet_bits_orred = 0
    for index, row in data.iterrows():
        fet_bits_orred = fet_bits_orred | row['balance_list']
    csv_data.insert(17, fet_bits_orred)    

    csv_data.insert(18, data['cell_count'].max())    
    csv_data.insert(19, data['ntc_count'].max())    
    csv_data.insert(20, data['temp_1'].max())    
    csv_data.insert(21, data['temp_1'].min())    
    csv_data.insert(22, data['temp_2'].max())    
    csv_data.insert(23, data['temp_2'].min())    
    csv_data.insert(24, data['power'].max())
    csv_data.insert(25, data['power'].mean())        
    csv_data.insert(26, data['voltage_01'].max())
    csv_data.insert(27, data['voltage_01'].min())
    csv_data.insert(28, data['voltage_02'].max())    
    csv_data.insert(29, data['voltage_02'].min())    
    csv_data.insert(30, data['voltage_03'].max())    
    csv_data.insert(31, data['voltage_03'].min())    
    csv_data.insert(32, data['voltage_04'].max())    
    csv_data.insert(33, data['voltage_04'].min())
    
    # Now we can write the interesting values to the csv file (append)
    file_exist = os.path.exists(obj['csv_bms_bigdata_file'])
    with open(obj['csv_bms_bigdata_file'], 'a', encoding='UTF8') as f:

        writer = csv.writer(f)

        # Add header when file is new
        if not file_exist:
            writer.writerow(["system_datetime", "total_voltage (max)", "total_voltage (min)", "current (max)", "current (min)", "remain_cap (max)", "remain_cap (min)", "delta_cap", "typ_cap (max)", "typ_cap (min)", "cycles", "prod_date", "balance_list", "prot_status", "softwar_version", "remain_cap_percent (max)", "remain_cap_percent (min)", "fet_bits", "cell_count", "ntc_count", "temp_1 (max)", "temp_1 (min)", "temp_2 (max)", "temp_2 (min)", "power (max)", "power (mean)", "voltage_01 (max)", "voltage_01 (min)", "voltage_02 (max)", "voltage_02 (min)", "voltage_03 (max)", "voltage_03 (min)", "voltage_04 (max)", "voltage_04 (min)"])

        # write the data
        writer.writerow(csv_data)

    # open Solar data file and extract values
    try:
        data = pd.read_csv(obj['csv_solar_data_file'])
    except FileNotFoundError:
        print("Input file <"+ obj['csv_solar_data_file'] +"> not found.")
        sys.exit();

#system_datetime,solar_voltage,solar_current,solar_power,load_voltage,load_current,load_power,battery_voltage,battery_current,battery_power,controller_temp,ambient_temp,generated_energy_today,controller_date_time,battery_status,controller_status,load_on

    iso_date = datetime.fromisoformat(data.iloc[0]['system_datetime'])
    
    csv_data = []
    csv_data.insert(0,iso_date.date())
    csv_data.insert(1, data['solar_voltage'].max())
    csv_data.insert(2, data['solar_voltage'].min())
    csv_data.insert(3, data['solar_current'].max())
    csv_data.insert(4, data['solar_current'].min())
    csv_data.insert(5, data['solar_power'].max())
    csv_data.insert(6, data['solar_power'].min())
    csv_data.insert(7, data['load_voltage'].max())
    csv_data.insert(8, data['load_voltage'].min())
    csv_data.insert(9, data['load_current'].max())
    csv_data.insert(10, data['load_current'].min())
    csv_data.insert(11, data['load_power'].max())
    csv_data.insert(12, data['load_power'].min())
    csv_data.insert(13, data['battery_voltage'].max())
    csv_data.insert(14, data['battery_voltage'].min())
    csv_data.insert(15, data['battery_current'].max())
    csv_data.insert(16, data['battery_current'].min())
    csv_data.insert(17, data['battery_power'].max())
    csv_data.insert(18, data['battery_power'].min())
    csv_data.insert(19, data['controller_temp'].max())
    csv_data.insert(20, data['controller_temp'].min())
    csv_data.insert(21, data['ambient_temp'].max())
    csv_data.insert(22, data['ambient_temp'].min())
    csv_data.insert(23, data['generated_energy_today'].max())
    
    iso_date = datetime.fromisoformat(data.iloc[0]['controller_date_time'])
    csv_data.insert(24, iso_date.date())
    
    # Get all bits from today
    battery_status_orred = 0
    for index, row in data.iterrows():
        battery_status_orred = battery_status_orred | row['battery_status']
    csv_data.insert(25, battery_status_orred)    

    # Get all bits from today
    controller_status_orred = 0
    for index, row in data.iterrows():
        controller_status_orred = controller_status_orred | row['controller_status']
    csv_data.insert(26, controller_status_orred)    

    # Now we can write the interesting values to the csv file (append)
    file_exist = os.path.exists(obj['csv_solar_bigdata_file'])
    with open(obj['csv_solar_bigdata_file'], 'a', encoding='UTF8') as f:

        writer = csv.writer(f)

        # Add header when file is new
        if not file_exist:
            writer.writerow(["system_datetime", "solar_voltage (max)", "solar_voltage (min)", "solar_current (max)", "solar_current (min)", "solar_power (max)", "solar_power (min)", "load_voltage (max)", "load_voltage (min)", "load_current (max)", "load_current (min)", "load_power (max)", "load_power (min)", "battery_voltage (max)", "battery_voltage (min)", "battery_current (max)", "battery_current (min)", "battery_power (max)", "battery_power (min)", "controller_temp (max)", "controller_temp (min)", "ambient_temp (max)", "ambient_temp (min)", "generated_energy_today", "controller_date_time", "battery_status", "controller_status"])

        # write the data
        writer.writerow(csv_data)
